/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package oracle.toplink.essentials.internal.helper;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.lang.ref.WeakReference;
import java.util.Enumeration;
import java.util.NoSuchElementException;

/**
 *
 * @author adam-bien.com
 */
public class WeakKeyAndValueIdentityHashtable extends IdentityHashtable{
       static final long serialVersionUID = 1421746759512286392L;
    
    // the default initial capacity
    static final int DEFAULT_INITIAL_CAPACITY = 32;

    // the maximum capacity.
    static final int MAXIMUM_CAPACITY = 1 << 30;

    // the loadFactor used when none specified in constructor.
    static final float DEFAULT_LOAD_FACTOR = 0.75f;

    /** An "enum" of Enumeration types. */
    static final int KEYS = 0;

    /** An "enum" of Enumeration types. */
    static final int ELEMENTS = 1;
    private static EmptyEnumerator emptyEnumerator = new EmptyEnumerator();
    protected transient WeakKeyEntry[] entries;// internal array of Entry's
    protected transient int count = 0;
    protected int threshold = 0;
    protected float loadFactor = 0;

    /**
     * Constructs a new <tt>HardIdentityHashtable</tt> with the given
     * initial capacity and the given loadFactor.
     *
     * @param initialCapacity the initial capacity of this
     * <tt>HardIdentityHashtable</tt>.
     * @param loadFactor the loadFactor of the <tt>HardIdentityHashtable</tt>.
     * @throws IllegalArgumentException  if the initial capacity is less
     * than zero, or if the loadFactor is nonpositive.
     */
    public WeakKeyAndValueIdentityHashtable(int initialCapacity, float loadFactor) {
        if (initialCapacity < 0) {
            throw new IllegalArgumentException("Illegal initialCapacity: " + initialCapacity);
        }
        if (initialCapacity > MAXIMUM_CAPACITY) {
            initialCapacity = MAXIMUM_CAPACITY;
        }
        if ((loadFactor <= 0) || Float.isNaN(loadFactor)) {
            throw new IllegalArgumentException("Illegal loadFactor: " + loadFactor);
        }

        // Find a power of 2 >= initialCapacity
        int capacity = 1;
        while (capacity < initialCapacity) {
            capacity <<= 1;
        }
        this.loadFactor = loadFactor;
        threshold = (int)(capacity * loadFactor);
        entries = new WeakKeyEntry[capacity];
    }

   /**
     * Constructs a new <tt>HardIdentityHashtable</tt> with the given
     * initial capacity and a default loadFactor of <tt>0.75</tt>.
     *
     * @param initialCapacity the initial capacity of the
     * <tt>HardIdentityHashtable</tt>.
     * @throws <tt>IllegalArgumentException</tt> if the initial capacity is less
     * than zero.
     */
    public WeakKeyAndValueIdentityHashtable(int initialCapacity) {
        this(initialCapacity, DEFAULT_LOAD_FACTOR);
    }

    /**
     * Constructs a new <tt>HardIdentityHashtable</tt> with a default initial
     * capacity of <tt>32</tt> and a loadfactor of <tt>0.75</tt>.
     */
    public WeakKeyAndValueIdentityHashtable() {
        loadFactor = DEFAULT_LOAD_FACTOR;
        threshold = (int)(DEFAULT_INITIAL_CAPACITY * DEFAULT_LOAD_FACTOR);
        entries = new WeakKeyEntry[DEFAULT_INITIAL_CAPACITY];
    }
    
    /**
     * Removes all of the mappings from this <tt>HardIdentityHashtable</tt>.
     */
    public synchronized void clear() {
        if (count > 0) {
            WeakKeyEntry[] copyOfEntries = entries;
            for (int i = copyOfEntries.length; --i >= 0;) {
                copyOfEntries[i] = null;
            }
            count = 0;
        }
    }

    /**
     * Returns a shallow copy of this <tt>HardIdentityHashtable</tt> (the
     * elements are not cloned).
     *
     * @return a shallow copy of this <tt>HardIdentityHashtable</tt>.
     */
    public synchronized Object clone() {
            WeakKeyEntry[] copyOfEntries = entries;
            WeakKeyAndValueIdentityHashtable clone = (WeakKeyAndValueIdentityHashtable)super.clone();
            clone.entries = new WeakKeyEntry[copyOfEntries.length];
            for (int i = copyOfEntries.length; i-- > 0;) {
                clone.entries[i] = (copyOfEntries[i] != null) ? (WeakKeyEntry)copyOfEntries[i].clone() : null;
            }
            return clone;
    }

    /**
     * Returns <tt>true</tt> if this <tt>HardIdentityHashtable</tt> contains
     * the given object. Equality is tested by the equals() method.
     *
     * @param obj the object to find.
     * @return <tt>true</tt> if this <tt>HardIdentityHashtable</tt> contains
     * obj.
     * @throws <tt>NullPointerException</tt> if obj is null</tt>.
     */
    public synchronized boolean contains(Object obj) {
        if (obj == null) {
            throw new NullPointerException();
        }

        WeakKeyEntry[] copyOfEntries = entries;
        for (int i = copyOfEntries.length; i-- > 0;) {
            for (WeakKeyEntry e = copyOfEntries[i]; e != null; e = e.next) {
                if (e.getValue().equals(obj)) {
                    return true;
                }
            }
        }
        return false;
    }

    /**
     * Returns <tt>true</tt> if this <tt>HardIdentityHashtable</tt> contains a
     * mapping for the given referenceOrKey. Equality is tested by reference.
     *
     * @param referenceOrKey object to be used as a referenceOrKey into this
     * <tt>HardIdentityHashtable</tt>.
     * @return <tt>true</tt> if this <tt>HardIdentityHashtable</tt> contains a
     * mapping for referenceOrKey.
     */
    public synchronized boolean containsKey(Object key) {
        WeakKeyEntry[] copyOfEntries = entries;
        int hash = System.identityHashCode(key);
        int index = (hash & 0x7FFFFFFF) % copyOfEntries.length;
        for (WeakKeyEntry e = copyOfEntries[index]; e != null; e = e.next) {
            if (e.getKey() == key) {
                return true;
            }
        }
        return false;
    }

    public synchronized Enumeration elements() {
        if (count == 0) {
            return emptyEnumerator;
        } else {
            return new Enumerator(ELEMENTS);
        }
    }

    /**
     * Returns the value to which the given referenceOrKey is mapped in this
     * <tt>HardIdentityHashtable</tt>. Returns <tt>null</tt> if this
     * <tt>HardIdentityHashtable</tt> contains no mapping for this referenceOrKey.
     *
     * @return the value to which this <tt>HardIdentityHashtable</tt> maps the
     * given referenceOrKey.
     * @param referenceOrKey referenceOrKey whose associated value is to be returned.
     */
    public synchronized Object get(Object key) {
        WeakKeyEntry[] copyOfEntries = entries;
        int hash = System.identityHashCode(key);
        int index = (hash & 0x7FFFFFFF) % copyOfEntries.length;
        for (WeakKeyEntry e = copyOfEntries[index]; e != null; e = e.next) {
            if (e.getKey() == key) {
                return e.getValue();
            }
        }
        return null;
    }

    /**
     * @return <tt>true</tt> if this <tt>HardIdentityHashtable</tt> is empty.
     */
    public boolean isEmpty() {
        return (count == 0);
    }

    public synchronized Enumeration keys() {
        if (count == 0) {
            return emptyEnumerator;
        } else {
            return new Enumerator(KEYS);
        }
    }

    /**
     * Associate the given object with the given referenceOrKey in this
     * <tt>HardIdentityHashtable</tt>, replacing any existing mapping.
     *
     * @param referenceOrKey referenceOrKey to map to given object.
     * @param obj object to be associated with referenceOrKey.
     * @return the previous object for referenceOrKey or <tt>null</tt> if this
     * <tt>HardIdentityHashtable</tt> did not have one.
     * @throws <tt>NullPointerException</tt> if obj is null</tt>.
     */
    public synchronized Object put(Object key, Object obj) {
        if (obj == null) {
            throw new NullPointerException();
        }

        WeakKeyEntry[] copyOfEntries = entries;
        int hash = System.identityHashCode(key);
        int index = (hash & 0x7FFFFFFF) % copyOfEntries.length;
        for (WeakKeyEntry e = copyOfEntries[index]; e != null; e = e.next) {
            if (e.getKey() == key) {
                Object old = e.getValue();
                e.setValue(obj);
                return old;
            }
        }

        if (count >= threshold) {
            rehash();
            copyOfEntries = entries;
            index = (hash & 0x7FFFFFFF) % copyOfEntries.length;
        }
        WeakKeyEntry e = new WeakKeyEntry(hash, key, obj, copyOfEntries[index]);
        copyOfEntries[index] = e;
        count++;
        return null;
    }

    /**
     * INTERNAL:
     * Re-builds the internal array of WeakKeyEntry's with a larger capacity.
     * This method is called automatically when the number of objects in this
     * HardIdentityHashtable exceeds its current threshold.
     */
    private void rehash() {
        int oldCapacity = entries.length;
        WeakKeyEntry[] oldEntries = entries;
        int newCapacity = (oldCapacity * 2) + 1;
        WeakKeyEntry[] newEntries = new WeakKeyEntry[newCapacity];
        threshold = (int)(newCapacity * loadFactor);
        entries = newEntries;
        for (int i = oldCapacity; i-- > 0;) {
            for (WeakKeyEntry old = oldEntries[i]; old != null;) {
                WeakKeyEntry e = old;
                old = old.next;
                int index = (e.hash & 0x7FFFFFFF) % newCapacity;
                e.next = newEntries[index];
                newEntries[index] = e;
            }
        }
    }

    /**
     * Removes the mapping (referenceOrKey and its corresponding value) from this
     * <tt>HardIdentityHashtable</tt>, if present.
     *
     * @param referenceOrKey referenceOrKey whose mapping is to be removed from the map.
     * @return the previous object for referenceOrKey or <tt>null</tt> if this
     * <tt>HardIdentityHashtable</tt> did not have one.
     */
    public synchronized Object remove(Object key) {
        WeakKeyEntry[] copyOfEntries = entries;
        int hash = System.identityHashCode(key);
        int index = (hash & 0x7FFFFFFF) % copyOfEntries.length;
        for (WeakKeyEntry e = copyOfEntries[index], prev = null; e != null; prev = e, e = e.next) {
            if (e.getKey() == key) {
                if (prev != null) {
                    prev.next = e.next;
                } else {
                    copyOfEntries[index] = e.next;
                }
                count--;
                return e.getValue();
            }
        }
        return null;
    }

    /**
     * @return the size of this <tt>HardIdentityHashtable</tt>.
     */
    public int size() {
        return count;
    }

    /**
     * Return the string representation of this <tt>HardIdentityHashtable</tt>.
     *
     * @return the string representation of this <tt>HardIdentityHashtable</tt>.
     */
    public synchronized String toString() {
        int max = size() - 1;
        StringBuffer buf = new StringBuffer();
        Enumeration k = keys();
        Enumeration e = elements();
        buf.append("{");
        for (int i = 0; i <= max; i++) {
            String s1 = k.nextElement().toString();
            String s2 = e.nextElement().toString();
            buf.append(s1 + "=" + s2);
            if (i < max) {
                buf.append(", ");
            }
        }
        buf.append("}");
        return buf.toString();
    }

    /**
     * Serialize the state of this <tt>HardIdentityHashtable</tt> to a stream.
     *
     * @serialData The <i>capacity</i> of the <tt>HardIdentityHashtable</tt>
     * (the length of the bucket array) is emitted (int), followed by the
     * <i>size</i> of the <tt>HardIdentityHashtable</tt>, followed by the
     * referenceOrKey-value mappings (in no particular order).
     */
    private void writeObject(ObjectOutputStream s) throws IOException {
        // Write out the threshold, loadfactor (and any hidden 'magic' stuff).
        s.defaultWriteObject();

        // Write out number of buckets
        s.writeInt(entries.length);
        // Write out count
        s.writeInt(count);
        // Write out contents
        for (int i = entries.length - 1; i >= 0; i--) {
            WeakKeyEntry entry = entries[i];
            while (entry != null) {
                s.writeObject(entry.getKey());
                s.writeObject(entry.getValue());
                entry = entry.next;
            }
        }
    }

    /**
     * Deserialize the <tt>HardIdentityHashtable</tt> from a stream.
     */
    private void readObject(ObjectInputStream s) throws IOException, ClassNotFoundException {
        // Read in the threshold, loadfactor (and any hidden 'magic' stuff).
        s.defaultReadObject();

        // Read in number of buckets and allocate the bucket array;
        int numBuckets = s.readInt();
        entries = new WeakKeyEntry[numBuckets];
        // Read in size (count)
        int size = s.readInt();

        // Read the mappings and add to the TopLinkIdentityHashMap
        for (int i = 0; i < size; i++) {
            Object key = s.readObject();
            Object value = s.readObject();
            put(key, value);
        }
    }

    private static class EmptyEnumerator implements Enumeration {
        EmptyEnumerator() {
        }

        public boolean hasMoreElements() {
            return false;
        }

        public Object nextElement() {
            throw new NoSuchElementException();
        }
    }

    class Enumerator implements Enumeration {
        int enumeratorType;
        int index;
        WeakKeyEntry entry;

        Enumerator(int enumeratorType) {
            this.enumeratorType = enumeratorType;
            index = WeakKeyAndValueIdentityHashtable.this.entries.length;
        }

        public boolean hasMoreElements() {
            if (entry != null) {
                return true;
            }
            while (index-- > 0) {
                if ((entry = WeakKeyAndValueIdentityHashtable.this.entries[index]) != null) {
                    return true;
                }
            }
            return false;
        }

        public Object nextElement() {
            if (entry == null) {
                while ((index-- > 0) && ((entry = WeakKeyAndValueIdentityHashtable.this.entries[index]) == null)) {
                    ;
                }
            }
            if (entry != null) {
                WeakKeyEntry e = entry;
                entry = e.next;
                if (enumeratorType == KEYS) {
                    return e.getKey();
                } else {
                    return e.getValue();
                }
            }
            throw new NoSuchElementException("IdentityHashtable.Enumerator");
        }
    }

    
    private static class WeakKeyEntry{
        int hash;
        private Object referenceOrKey;
        Object value;
        WeakKeyEntry next;

        WeakKeyEntry(int hash, Object key, Object value, WeakKeyEntry next) {
            this.hash = hash;
            this.referenceOrKey = wrap(key);
            this.value = value;
            this.next = next;
        }

        protected Object clone() {
            return new WeakKeyEntry(hash, unwrap(referenceOrKey), value, ((next == null) ? null : (WeakKeyEntry)next.clone()));
        }

        public Object getKey() {
            return unwrap(referenceOrKey);
        }

        public Object getValue() {
            return value;
        }

        public Object setValue(Object value) {
            Object oldValue = this.value;
            this.value = value;
            return oldValue;
        }

        public boolean equals(Object o) {
            if (!(o instanceof WeakKeyEntry)) {
                return false;
            }

            WeakKeyEntry e = (WeakKeyEntry)o;
            return (referenceOrKey == unwrap(e.getKey()) && (value == null) ? (e.getValue() == null) : value.equals(e.getValue()));
        }

        public int hashCode() {
            return hash ^ ((value == null) ? 0 : value.hashCode());
        }

        public String toString() {
            return referenceOrKey + "=" + unwrap(value);
        }
        
        private Object unwrap(Object referent){
            WeakReference wrapper = (WeakReference) referent;
            return wrapper.get();
        }
        
        private Object wrap(Object referent){
            //already wrapped, nothing to do...
            if(referent instanceof WeakReference){
                return referent;
            }
            return  new WeakReference(referent);
        }
    }
    

}
